"
 COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
"{ Package: 'stx:libbasic' }"

SequenceableCollection subclass:#LinkedList
	instanceVariableNames:'firstLink lastLink numberOfNodes'
	classVariableNames:''
	poolDictionaries:''
	category:'Collections-Linked'
!

!LinkedList class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

documentation
"
    this class implements an anchor to a list of Links.
    The data itself is held in the link elements.
    See (the abstract) Link, ValueLink and (possibly other) classes,
    which can be used as elements of a linkedList.

    LinkedList does not care for storage; all it does is handling
    chained link elements, which must respond to #nextLink/#nextLink:.
    (i.e. any object which can do this, can be used as elements of a linked
     list).
    An abstract superclass for linkElements is Link; a concrete class is
    ValueLink, which holds a reference to some object.

    [warning:]
        Be careful when subclassing Link, since there is a big drawback,
        which may be overlooked by beginners:
            a Link element can only be in one LinkedList
            - adding the same element to another LinkedList
            will remove it from the first as a side effect.
        Therefore, NEVER simply add something to a linkedList (except for
        valueLinks) unless you know what you do.
        The ST-80 implementors probably wanted this behavior, to move
        processes from the waitingList to runLists and vice versa;
        however, literature seems to not point this out enough.

    Although LinkedList is a subclass of SequenceableCollection (and therefore
    supports indexed access via at:), you should be careful in using it or
    other methods based upon at:.
    The reason is that #at: walks the linkedlist to find the indexed element
    and is therefore slow.
    This means that some linear-in-time algorithms inherited from
    SequenceableCollection become square in runtime.
    In general, if you need access via a numeric index, you better use Array,
    OrderedCollection or similar.

    For the above reasons, the system does not make heavily use of LinkedLists;
    the only good application is where elements must be repeatedly be removed
    at the front and added at the end.
    (the schedulers process handling code does this to manage process lists.)

    [memory requirements:]
        (OBJ-HEADER + (3 * ptr-size)) * size
                    + any additional instvars due to subclassing

    [author:]
        Claus Gittinger (July 1993)

    [see also:]
        Link ValueLink Process
"
!

examples
"
									[exBegin]
    |l|

    l := LinkedList new.
    l addLast:(ValueLink new value:'one').
    l addLast:(ValueLink new value:'two').
    l addLast:(ValueLink new value:'three').
    l addLast:(ValueLink new value:'four').
    l inspect
									[exEnd]


									[exBegin]
    |l|

    l := LinkedList new.
    l addLast:(ValueLink new value:'one').
    l addLast:(ValueLink new value:'two').
    l addLast:(ValueLink new value:'three').
    l addLast:(ValueLink new value:'four').
    (l at:3) value inspect.        'slow operation for large lists'.
									[exEnd]


									[exBegin]
    |l link|

    l := LinkedList new.
    l addLast:(ValueLink new value:'one').
    l addLast:(ValueLink new value:'two').
    l addLast:(ValueLink new value:'three').
    l addLast:(ValueLink new value:'four').
    link := l removeFirst.
    l addLast:link.
    l inspect.
									[exEnd]
"
! !

!LinkedList class methodsFor:'instance creation'!

new
    "create and return a new LinkedList"

    ^ self basicNew initialize
! !

!LinkedList methodsFor:'accessing'!

at:index
    "return the n'th element - use of this method should be avoided,
     since it is slow to walk through the list - think about using
     another collection if you need indexed access.
     Notice, that many methods in SeqColl are based on at:-access,
     so other inherited methods may be very slow (showing square runtime)."

    ^ self at:index ifAbsent:[ self subscriptBoundsError:index]
!

at:index ifAbsent:exceptionBlock
    "return the n'th element - use of this method should be avoided,
     since it is slow to walk through the list - think about using
     another collection if you need indexed access.
     Notice, that many methods in SeqColl are based on at:-access,
     so other inherited methods may be very slow (showing square runtime)."

    |theLink
     runIndex "{Class: SmallInteger}"|

    theLink := firstLink.
    runIndex := 1.
    [runIndex == index] whileFalse:[
	theLink isNil ifTrue:[^ exceptionBlock value].
	theLink := theLink nextLink.
	runIndex := runIndex + 1.
    ].
    ^ theLink
!

first
    "return the first node in the list"

    firstLink isNil ifTrue:[^ self emptyCollectionError].
    ^ firstLink
!

firstIfEmpty:exceptionalValue
    "return the first node in the list or exceptionlValue, if empty"

    firstLink isNil ifTrue:[^ exceptionalValue value].
    ^ firstLink
!

last
    "return last node in the list"

    lastLink isNil ifTrue:[self emptyCollectionError].
    ^ lastLink
! !

!LinkedList methodsFor:'adding & removing'!

add:aLink
    "adds aLink to the end of the sequence. Returns aLink"

    aLink nextLink:nil.
    lastLink isNil ifTrue:[
	firstLink := aLink
    ] ifFalse: [
	lastLink nextLink:aLink
    ].
    lastLink := aLink.
    numberOfNodes := numberOfNodes + 1.
    ^ aLink
!

add:linkToAdd after:aLink
    "adds linkToAdd after another link, aLink. If aLink is nil,
     linkToAdd is inserted at the beginning. Returns linkToAdd."

    |this|

    aLink isNil ifTrue:[^ self addFirst:linkToAdd ].

    this := firstLink.
    [this notNil and:[this ~~ aLink]] whileTrue:[
	this := this nextLink
    ].
    this isNil ifTrue:[^ self add:linkToAdd ].
    linkToAdd nextLink:(this nextLink).
    this nextLink:linkToAdd.
    numberOfNodes := numberOfNodes + 1.
    ^ linkToAdd
!

addFirst:aLink
    "adds aLink to the beginning of the sequence. Returns aLink"

    firstLink isNil ifTrue:[
	lastLink := aLink.
    ].
    aLink nextLink:firstLink.
    firstLink := aLink.
    numberOfNodes := numberOfNodes + 1.
    ^ aLink
!

remove:aLink ifAbsent:exceptionBlock
    "remove the argument, aLink from the sequence and return it;
     if absent, evaluate the exceptionBlock.
     Actually this is really a #removeIdentical (but for compatibility...)"

    ^ self removeIdentical:aLink ifAbsent:exceptionBlock

    "Modified: / 30-11-2010 / 13:38:45 / cg"
!

removeAll
    "remove all elements from the sequence. Returns the receiver."

    firstLink := lastLink := nil.
    numberOfNodes := 0

    "Created: 21.3.1996 / 15:24:38 / cg"
    "Modified: 12.4.1996 / 13:34:53 / cg"
!

removeFirst
    "remove and return the first node from the sequence"

    |link|

    firstLink isNil ifTrue:[
	^ self emptyCollectionError
    ].
    link := firstLink.
    firstLink := firstLink nextLink.
    firstLink isNil ifTrue:[
	lastLink := nil
    ].
    link nextLink:nil.
    numberOfNodes := numberOfNodes - 1.
    ^ link
!

removeIdentical:aLink ifAbsent:exceptionBlock
    "remove the argument, aLink from the sequence and return it;
     if absent, evaluate the exceptionBlock."

    |prevNode nextNode thisNode|

    thisNode := firstLink.
    [thisNode notNil] whileTrue:[
        nextNode := thisNode nextLink.
        (thisNode == aLink) ifTrue:[
            prevNode isNil ifTrue:[
                firstLink := nextNode
            ] ifFalse:[
                prevNode nextLink:nextNode
            ].
            nextNode isNil ifTrue:[
                lastLink := prevNode
            ].
            numberOfNodes := numberOfNodes - 1.
            thisNode nextLink:nil.
            ^ aLink
        ].
        prevNode := thisNode.
        thisNode := nextNode
    ].
    ^ exceptionBlock value

    "Created: / 30-11-2010 / 13:38:25 / cg"
! !

!LinkedList methodsFor:'enumerating'!

do:aBlock
    "evaluate the argument, aBlock with 1 arg for every element in the list"

    |thisNode|

    thisNode := firstLink.
    [thisNode notNil] whileTrue:[
	aBlock value:thisNode.
	thisNode := thisNode nextLink
    ]
! !

!LinkedList methodsFor:'initialization'!

initialize
    numberOfNodes := 0
! !

!LinkedList methodsFor:'queries'!

size
    "return the size of the LinkedList i.e. the number of nodes"

    ^ numberOfNodes
! !

!LinkedList methodsFor:'testing'!

identityIndexOf:aLink startingAt:start
    "search the collection for aLink, starting the search at index start;
     if found, return the index otherwise return 0. Here, index is defined
     as the link-nodes position in the list.
     The comparison is done using ==
     (i.e. identity test - not equality test)."

    |theNode idx "{ Class: SmallInteger }"|

    theNode := firstLink.
    idx := 1.
    [idx < start] whileTrue:[
        theNode isNil ifTrue:[^ 0].     "reached the end"
        theNode := theNode nextLink.
        idx := idx + 1.
    ].
    [theNode notNil] whileTrue:[
        (aLink == theNode) ifTrue:[^ idx].
        theNode := theNode nextLink.
        idx := idx + 1.
    ].                                  "reached the end"
    ^ 0

    "
     |l|

     l := LinkedList new.
     l indexOf:'hello'
    "

    "
     |l v|

     l := LinkedList new.
     l add:(ValueLink new value:'one').
     l add:(ValueLink new value:'two').
     l add:(v := ValueLink new value:'hello').
     l identityIndexOf:v
    "
!

indexOf:aLink startingAt:start
    "search the collection for aLink, starting the search at index start;
     if found, return the index otherwise return 0. Here, index is defined
     as the link-nodes position in the list.
     The comparison is done using = (i.e. equality test - not identity test)."

    |theNode idx "{ Class: SmallInteger }"|

    theNode := firstLink.
    idx := 1.
    [idx < start] whileTrue:[
	theNode isNil ifTrue:[^ 0].     "reached the end"
	theNode := theNode nextLink.
	idx := idx + 1.
    ].
    [theNode notNil] whileTrue:[
	(aLink = theNode) ifTrue:[^ idx].
	theNode := theNode nextLink.
	idx := idx + 1.
    ].                                  "reached the end"
    ^ 0

    "
     |l|

     l := LinkedList new.
     l indexOf:'hello'
    "

    "
     |l v|

     l := LinkedList new.
     l add:(ValueLink new value:'one').
     l add:(ValueLink new value:'two').
     l add:(v := ValueLink new value:'hello').
     l indexOf:v
    "
!

isEmpty
    "return true, if the collection is empty"

    ^ firstLink isNil
!

notEmpty
    "return true, if the collection is not empty"

    ^ firstLink notNil
! !

!LinkedList class methodsFor:'documentation'!

version
    ^ '$Id: LinkedList.st 10695 2011-09-28 19:07:49Z vranyj1 $'
!

version_CVS
    ^ 'Header: /cvs/stx/stx/libbasic/LinkedList.st,v 1.41 2011/09/27 09:42:48 cg Exp '
!

version_SVN
    ^ '$Id: LinkedList.st 10695 2011-09-28 19:07:49Z vranyj1 $'
! !
